//                                               -*- C++ -*-
/**
 *  @file  GramSchmidtAlgorithm.cxx
 *  @brief OrthogonalUniVariatePolynomialStandardDistribution polynomial factory
 *
 *  (C) Copyright 2005-2010 EDF-EADS-Phimeca
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  @author: $LastChangedBy: dutka $
 *  @date:   $LastChangedDate: 2008-05-21 11:21:38 +0200 (Wed, 21 May 2008) $
 *  Id:      $Id: Object.cxx 815 2008-05-21 09:21:38Z dutka $
 */

#include "GramSchmidtAlgorithm.hxx"
#include "PersistentObjectFactory.hxx"
#include "Uniform.hxx"
#include "Exception.hxx"
#include "SquareMatrix.hxx"
#include "Log.hxx"

namespace OpenTURNS
{

  namespace Uncertainty
  {

    namespace Algorithm
    {

      CLASSNAMEINIT(GramSchmidtAlgorithm);

      static Base::Common::Factory<GramSchmidtAlgorithm> RegisteredFactory("GramSchmidtAlgorithm");

      typedef Distribution::Uniform                  Uniform;
      typedef Base::Common::InvalidArgumentException InvalidArgumentException;

      /* Default constructor */
      GramSchmidtAlgorithm::GramSchmidtAlgorithm()
        : OrthonormalizationAlgorithmImplementation(Uniform()),
          standardMoments_(1, 1.0),
          coefficientsCache_(1, Coefficients(1, 1.0))
      {
        // Nothing to do
      }


      /* Parameter constructor */
      GramSchmidtAlgorithm::GramSchmidtAlgorithm(const Distribution & measure)
        : OrthonormalizationAlgorithmImplementation(measure),
          standardMoments_(1, 1.0),
          coefficientsCache_(1, Coefficients(1, 1.0))
      {
        // Nothing to do
      }


      /* Virtual constructor */
      GramSchmidtAlgorithm * GramSchmidtAlgorithm::clone() const
      {
        return new GramSchmidtAlgorithm(*this);
      }


      /* Calculate the coefficients of recurrence a0n, a1n, a2n such that
         Pn+1(x) = (a0n * x + a1n) * Pn(x) + a2n * Pn-1(x) */
      GramSchmidtAlgorithm::Coefficients GramSchmidtAlgorithm::getRecurrenceCoefficients(const UnsignedLong n) const
      {
        // An orthonormal polynomial of degree n writes P(x) = k_n * x^n + l_n * x^{n-1} + ...
        // Degree n+1
        const Coefficients coefficientsNp1(buildPolynomial(n + 1).getCoefficients());
        const NumericalScalar kNp1(coefficientsNp1[n + 1]);
        const NumericalScalar lNp1(coefficientsNp1[n]);
        // Degree n
        const Coefficients coefficientsN(buildPolynomial(n).getCoefficients());
        const NumericalScalar kN(coefficientsN[n]);
        NumericalScalar lN(0.0);
        NumericalScalar kNm1(0.0);
        if (n > 0)
          {
            // Degree n-1, not used for n = 0
            lN = coefficientsN[n - 1];
            const Coefficients coefficientsNm1(buildPolynomial(n - 1).getCoefficients());
            kNm1 = coefficientsNm1[n - 1];
          }
        Coefficients result(3);
        result[0] = kNp1 / kN;
        result[1] = result[0] * (lNp1 / kNp1 - lN / kN);
        result[2] = -result[0] * kNm1 / kN;
        return result;
      }


      /* Return the order-th raw moment of the underlying measure */
      NumericalScalar GramSchmidtAlgorithm::getStandardMoment(const UnsignedLong order) const
      {
        // We know that the raw moments will be accessed in a particular pattern: the moments not already
        // computed will always be accessed in a successive increasing order
        const UnsignedLong maxOrder(standardMoments_.getSize());
        if (order > maxOrder) throw InvalidArgumentException(HERE) << "Error: cannot access to the raw moments in arbitrary order.";
        if (order == maxOrder)
          standardMoments_.add(measure_.getStandardMoment(order)[0]);
        return standardMoments_[order];
      }

      /* Build the kth orthonormal polynomial */
      GramSchmidtAlgorithm::UniVariatePolynomial GramSchmidtAlgorithm::buildPolynomial(const UnsignedLong k) const
      {
        // If the needed polynomial is already in the cache
        if (k < coefficientsCache_.getSize()) return coefficientsCache_[k];
        if (k > coefficientsCache_.getSize()) throw InvalidArgumentException(HERE) << "Error: cannot call buildPolynomial in arbitrary order.";
        // Build v = x^k
        Coefficients coefficients(k + 1, 0.0);
        coefficients[k] = 1.0;
        UniVariatePolynomial v(coefficients);
        // Modified Gram-Schmidt algorithm
        UniVariatePolynomial q;
        for (UnsignedLong i = 0; i < k; ++i)
          {
            UniVariatePolynomial qi(buildPolynomial(i));
            v = v - qi * dotProduct(qi, v);
            q = v * (1.0 / sqrt(dotProduct(v, v)));
          }
        coefficientsCache_.add(q.getCoefficients());
        return q;
      }


      /* Compute the dot product between two general polynomials according to the measure */
      NumericalScalar GramSchmidtAlgorithm::dotProduct(const UniVariatePolynomial & p1,
                                                       const UniVariatePolynomial & p2) const
      {
        NumericalScalar value(0.0);
        // Use the Kahan compensated summation to reduce roundoff errors
        NumericalScalar e(0.0);
        const UniVariatePolynomial q(p1 * p2);
        const Coefficients ai(q.getCoefficients());
        const UnsignedLong dimension(ai.getDimension());
        for (UnsignedLong i = 0; i < dimension; ++i)
          {
            const NumericalScalar temp(value);
            const NumericalScalar y(ai[i] * getStandardMoment(i) + e);
            value = temp + y;
            e = (temp - value) + y;
          }
        return value + e;
      }


      /* String converter */
      String GramSchmidtAlgorithm::__repr__() const
      {
        return OSS() << "class=" << getClassName()
                     << " measure=" << measure_;
      }


    } /* namespace Algorithm */
  } /* namespace Uncertainty */
} /* namespace OpenTURNS */
