//                                               -*- C++ -*-
/**
 *  @file  SafeAndSlow.cxx
 *  @brief Find the roots in a given direction according to the Proban
 *
 *  (C) Copyright 2005-2010 EDF-EADS-Phimeca
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  @author: $LastChangedBy: dutka $
 *  @date:   $LastChangedDate: 2010-02-04 16:44:49 +0100 (jeu. 04 févr. 2010) $
 *  Id:      $Id: SafeAndSlow.cxx 1473 2010-02-04 15:44:49Z dutka $
 */
#include "SafeAndSlow.hxx"
#include "NumericalPoint.hxx"

namespace OpenTURNS
{

  namespace Uncertainty 
  {

    namespace Algorithm
    {

      typedef OT::Base::Type::NumericalPoint NumericalPoint;

      /**
       * @class SafeAndSlow
       */

      CLASSNAMEINIT(SafeAndSlow);

      /* Constructor with parameters */
      SafeAndSlow::SafeAndSlow():
	RootStrategyImplementation()
      {
	// Nothing to do
      }

      /* Constructor with parameters */
      SafeAndSlow::SafeAndSlow(const Solver & solver):
	RootStrategyImplementation(solver)
      {
	// Nothing to do
      }

      /* Constructor with parameters */
      SafeAndSlow::SafeAndSlow(const Solver & solver,
			       const NumericalScalar maximumDistance,
			       const NumericalScalar stepSize):
	RootStrategyImplementation(solver, maximumDistance, stepSize)
      {
	// Nothing to do
      }

      /* Virtual constructor */
      SafeAndSlow * SafeAndSlow::clone() const
      {
	return new SafeAndSlow(*this);
      }

      /* Solve gives all the roots found applying the root strategy */
      SafeAndSlow::NumericalScalarCollection SafeAndSlow::solve(const NumericalMathFunction & function,
								const NumericalScalar value)
      {
	NumericalScalarCollection result(0);
	NumericalScalar infPoint(0.0);
	NumericalScalar infValue;
	// Get the value of the function at the origin
	try
	  {
	    infValue = getOriginValue();
	  }
	// If it has not yet been computed, compute it and store it
	catch (NotDefinedException & ex)
	  {
	    infValue = function(NumericalPoint(1, infPoint))[0];
	    setOriginValue(infValue);
	  }
        NumericalScalar maximumDistance(getMaximumDistance());
        NumericalScalar stepSize(getStepSize());
	Solver solver(getSolver());
	while(infPoint < maximumDistance)
	  {
	    NumericalScalar supPoint(std::min(infPoint + stepSize, maximumDistance));
	    NumericalScalar supValue(function(NumericalPoint(1, supPoint))[0]);
	    if ((infValue - value) * (supValue - value) < 0.0)
	      {
		result.add(solver.solve(function, value, infPoint, supPoint, infValue, supValue));
	      }
	    infPoint = supPoint;
	    infValue = supValue;
	  }
	return result;
      }

      /* String converter */
      String SafeAndSlow::__repr__() const
      {
	OSS oss;
	oss << "class=" << SafeAndSlow::GetClassName()
	    << " derived from " << RootStrategyImplementation::__repr__();
	return oss;
      }

    } /* namespace Algorithm */
  } /* namespace Uncertainty */
} /* namespace OpenTURNS */
